//+------------------------------------------------------------------+
//|                                              qpu_EMA-Rainbow.mq5 |
//|                                       Copyright 2012, MetaDriver |
//|                                            MetaDriver@rambler.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2012, MetaDriver"
#property link      "MetaDriver@rambler.ru"
#property version   "1.00"
//---
#define  CountLine  64
//---
#property indicator_chart_window
#property indicator_buffers CountLine
#property indicator_plots   CountLine
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum clRealType
  {
   cl_Float,  // float
   cl_Double  // double
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum eFutureMode
  {
   Back,    //   ()
   Forward  //  () 
  };
//---  
input eFutureMode FutureMode   =Forward;          //   ;-)
input bool        UseCpu       =true;             //  CPU
input clRealType  cl_RType     =cl_Double;        // OpelCL RealType
input int         FastPeriod   =3;
input int         PeriodStep   =2;
input color       FastColor    =clrPaleGoldenrod;
input color       MiddleColor  =clrLightSalmon;
input color       SlowColor    =clrMaroon;
input color       LFastColor   =clrPaleVioletRed;
input color       LMiddleColor =clrDarkTurquoise;
input color       LSlowColor   =clrMaroon;
//--- 
struct SBuffer
  {
   double            B[];
  };
//--- indicator buffers
SBuffer       EmaBuffer[CountLine];
//--- Calc buffers  and other vars
float         gpu_float_Buffer[][CountLine];
double        gpu_double_Buffer[][CountLine];
//---
int           cl_BufLength=1024*128;
//---
color Colors[CountLine];
bool  First=true;
//+------------------------------------------------------------------+
//| Custom indicator initialization function                         |
//+------------------------------------------------------------------+
int OnInit()
  {
//--- indicator buffers mapping
   InitColors();
//---
   for(int i=0; i<CountLine; i++)
      InitBuffers(i);
//---
   return(0);
  }
//+------------------------------------------------------------------+
//| Custom indicator iteration function                              |
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,
                const int prev_calculated,
                const int begin,
                const double &price[])
  {
//---
   if(FutureMode==Back)
      ArraySetAsSeries(price,true);
//---
   int limit=rates_total-prev_calculated;
//---
   if(limit==1)
      limit++;
//---
   long StartTime=0;
//---
   if(First)
      StartTime=GetTickCount();
//---
   if(prev_calculated==0)
     {
      limit-=begin+1;
      for(int i=0;i<CountLine;i++)
         EmaBuffer[i].B[limit]=price[limit];
      //---      ... :)
      if(!UseCpu)
         CalcEMAsAtGpu(limit,price);
     }
//---
   if(limit!=0)
     {
      for(int i=0;i<CountLine;i++)
        {
         double p=2./(FastPeriod+PeriodStep*(CountLine-i)+1);
         //---
         for(int j=limit-1;j>=0;j--)
           {
            //EmaBuffer[i].B[j] = price[j]+(0.0001*i-0.0032)*5;
            EmaBuffer[i].B[j]=price[j]*p+EmaBuffer[i].B[j+1]*(1.-p);
           }
        }
     }
//---
   if(First)
     {
      Print("Calculate ",rates_total," bars at ",
            (UseCpu ? "CPU, " : "GPU (OpenCL), "),"time = ",GetTickCount()-StartTime," ms");
      First=false;
     }
//--- return value of prev_calculated for next call
   return(rates_total);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void InitColors()
  {
   int Half=CountLine/2;
//---
   for(int i=0; i<Half; i++)
     {
      if(i%4==3)
         Colors[i]=MixColor(LSlowColor,LMiddleColor,i*1.0/Half);
      else
         Colors[i]=MixColor(SlowColor,MiddleColor,i*1.0/Half);
     }
//---
   for(int i=Half; i<CountLine; i++)
     {
      if(i%4==3)
         Colors[i]=MixColor(LMiddleColor,LFastColor,(i-Half)*1.0/Half);
      else
         Colors[i]=MixColor(MiddleColor,FastColor,(i-Half)*1.0/Half);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
color MixColor(color A,color B,double mix)
  {
   int r = (int)MathRound((B     & 0xFF) * mix + (A     & 0xFF) * (1. - mix));
   int g = (int)MathRound((B>>8  & 0xFF) * mix + (A>>8  & 0xFF) * (1. - mix));
   int b = (int)MathRound((B>>16 & 0xFF) * mix + (A>>16 & 0xFF) * (1. - mix));
   return color(r|g<<8|b<<16);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void InitBuffers(int index)
  {
   SetIndexBuffer(index,EmaBuffer[index].B,INDICATOR_DATA);
   PlotIndexSetInteger(index,PLOT_DRAW_TYPE,DRAW_LINE);
   PlotIndexSetInteger(index,PLOT_LINE_COLOR,Colors[index]);
   PlotIndexSetDouble(index,PLOT_EMPTY_VALUE,EMPTY_VALUE);
//---
   if(index%8==7) PlotIndexSetInteger(index,PLOT_LINE_WIDTH,1);
   if(FutureMode==Back) ArraySetAsSeries(EmaBuffer[index].B,true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CalcEMAsAtGpu(int &limit,const double &price[])
  {
   ResetLastError();
//---
   int cl_context_handle =INVALID_HANDLE;
   int cl_program_handle =INVALID_HANDLE;
   int cl_kernel_handle  =INVALID_HANDLE;
   int cl_mem_in_handle         =0;
   int cl_mem_out_handle        =0;
//---
   if((cl_context_handle=CLContextCreate(CL_USE_GPU_ONLY))==INVALID_HANDLE)
     {
      Print("OpenCL not found. Using CPU for calculation.");
      return;
     }
//---
   if((cl_program_handle=CLProgramCreate(cl_context_handle,Get_cl_Source()))==INVALID_HANDLE)
     {
      CLContextFree(cl_context_handle);
      Print("OpenCL program create failed (error: "+i2s(GetLastError())+"). Using CPU for calculation.");
      return;
     }
//---
   if((cl_kernel_handle=CLKernelCreate(cl_program_handle,"gpu_Ema"))==INVALID_HANDLE)
     {
      CLProgramFree(cl_program_handle);
      CLContextFree(cl_context_handle);
      //---
      Print("OpenCL kernel create failed. Using CPU for calculation.");
      Print("GetLastError returned .. ",GetLastError());
      switch(GetLastError())
        {
         case ERR_OPENCL_INVALID_HANDLE:
            Print("Err ==    program OpenCL");
            break;
         case ERR_INVALID_PARAMETER:
            Print("Err ==   ");
            break;
         case ERR_OPENCL_TOO_LONG_KERNEL_NAME:
            Print("Err ==     127 ");
            break;
         case ERR_OPENCL_KERNEL_CREATE:
            Print("Err ==       OpenCL.");
            break;
         default: Print("Err ==  .");
        }
      return;
     }
//---
   if((cl_mem_in_handle=CLBufferCreate(cl_context_handle,
      cl_BufLength*((cl_RType==cl_Float)?sizeof(float):sizeof(double)),
      CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLKernelFree(cl_kernel_handle);
      CLProgramFree(cl_program_handle);
      CLContextFree(cl_context_handle);
      //---
      Print("OpenCL input buffer create failed. Using CPU for calculation.");
      return;
     }
//---
   if((cl_mem_out_handle=CLBufferCreate(cl_context_handle,
      cl_BufLength*CountLine*((cl_RType==cl_Float)?sizeof(float):sizeof(double)),
      CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLBufferFree(cl_mem_in_handle);
      CLKernelFree(cl_kernel_handle);
      CLProgramFree(cl_program_handle);
      CLContextFree(cl_context_handle);
      //---
      Print("OpenCL output buffer create failed. Using CPU for calculation.");
      return;
     }
//---
   float fIn[],fOut[];  //   -
   double dIn[],dOut[];
//---
   if(cl_RType==cl_Float)
     {
      ArrayResize(fIn,cl_BufLength);
      ArrayResize(fOut,cl_BufLength*CountLine);
     }
   else
     {
      ArrayResize(dIn,cl_BufLength);
      ArrayResize(dOut,cl_BufLength*CountLine);
     }
//---     OpenCL
   CLSetKernelArg(cl_kernel_handle,0,cl_BufLength);
   CLSetKernelArgMem(cl_kernel_handle,1,cl_mem_in_handle);
   CLSetKernelArgMem(cl_kernel_handle,2,cl_mem_out_handle);
//---
   uint kOfs[1]  ={0};
   uint kWork[1] ={CountLine};
//--- Main loop
   while(limit/(cl_BufLength-1)>0)
     {
      if(cl_RType==cl_Float)
        {
         //---    Ema
         for(int i=0; i<CountLine; i++)
            fOut[i]=(float)EmaBuffer[i].B[limit];
         //---   gpu
         CLBufferWrite(cl_mem_out_handle,fOut,0,CountLine*sizeof(float));
         //---  
         for(int i=1; i<cl_BufLength; i++)
            fIn[i]=(float)price[limit-i];
         //---   gpu
         CLBufferWrite(cl_mem_in_handle,fIn);
        }
      else // (cl_RType==cl_Double)
        {
         //---    Ema
         for(int i=0; i<CountLine; i++)
            dOut[i]=EmaBuffer[i].B[limit];
         //---   gpu
         CLBufferWrite(cl_mem_out_handle,dOut,0,CountLine*sizeof(double));
         //---  
         for(int i=1; i<cl_BufLength; i++)
            dIn[i]=price[limit-i];
         //---   gpu
         CLBufferWrite(cl_mem_in_handle,dIn);
        }
      //---
      CLExecute(cl_kernel_handle,1,kOfs,kWork);
      //---
      if(cl_RType==cl_Float)
        {
         CLBufferRead(cl_mem_out_handle,fOut);
         for(int i=1,r=limit-1; i<cl_BufLength; i++,r--)
           {
            for(int j=0; j<CountLine; j++)
               EmaBuffer[j].B[r]=fOut[i*CountLine+j];
           }
        }
      else
        {
         CLBufferRead(cl_mem_out_handle,dOut);
         for(int i=1,r=limit-1; i<cl_BufLength; i++,r--)
           {
            for(int j=0; j<CountLine; j++)
               EmaBuffer[j].B[r]=dOut[i*CountLine+j];
           }
        }
      //---
      limit-=(cl_BufLength-1);
     }
//---
   CLBufferFree(cl_mem_out_handle);
   CLBufferFree(cl_mem_in_handle);
   CLKernelFree(cl_kernel_handle);
   CLProgramFree(cl_program_handle);
   CLContextFree(cl_context_handle);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string i2s(int x)
  {
   return IntegerToString(x);
  }
//+------------------------------------------------------------------+
//|  OpenCL                                                       |
//+------------------------------------------------------------------+
string Get_cl_Source()
  {
   string cl_Source=((cl_RType==cl_Float)
                     ?
                     "#define real float                                                                    \r\n"
                     :
                     "#pragma OPENCL EXTENSION cl_khr_fp64 : enable                                         \r\n"
                     "#define real double                                                                   \r\n"
                     )+
                     "                                                                                      \r\n"
                     "__kernel void gpu_Ema(int CountCalc,                                                  \r\n"
                     "                      __global real *PriceBuffer,                                     \r\n"
                     "                      __global real *EmaBuffer)                                       \r\n"
                     "  {                                                                                   \r\n"
                     "   int  gID =get_global_id(0);                                                        \r\n"
                     "   real p   =2./("+i2s(FastPeriod)+"+"+i2s(PeriodStep)+"*("+i2s(CountLine)+"-gID)+1); \r\n"
                     "                                                                                      \r\n"
                     "   for(int i=1; i<CountCalc; i++)                                                     \r\n"
                     "     {                                                                                \r\n"
                     "      EmaBuffer[gID+"+i2s(CountLine)+"*i]=                                            \r\n"
                     "      EmaBuffer[gID+"+i2s(CountLine)+"*(i-1)]*(1.-p)+PriceBuffer[i]*p;                \r\n"
                     "     }                                                                                \r\n"
                     "  }                                                                                   \r\n";
   return(cl_Source);
  }
//+------------------------------------------------------------------+
